

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Enumeration;
import java.lang.Math;
import java.awt.TextArea;
import java.awt.Color;

public class Graph implements CommandExecutor
{

  private Hashtable nodes;
  private Vector faces;
  public Vector normals;  // the face normals computed are kept here
  private Vector retain;
  public Vector bestVectors;

  // needed by canvas
  public Hashtable projectedNodes;
  public Vector edges;
  public Node avgNode;
  public Node bestVect;
  public TextArea display;

  
  public Graph()
  {
    nodes = new Hashtable();
    edges = new Vector();
    faces = new Vector();
    normals = new Vector();
    retain = new Vector();
    bestVectors = new Vector();
    bestVect = null;
    avgNode = null;
    projectedNodes = new Hashtable();
  }

  public Graph(CommandParser parse, TextArea inDisplay)
  {
    nodes = new Hashtable();
    edges = new Vector();
    faces = new Vector();
    normals = new Vector();
    retain = new Vector();
    bestVectors = new Vector();
    bestVect = null;
    avgNode = null;
    projectedNodes = new Hashtable();

    parse.AddCommand("pt <str> <dbl>,<dbl>,<dbl>",this);
    parse.AddCommand("edge <str> <str>",this);


    // this is done to accept variable length arguments.  An anonymous subclass 
    // of command (instantiated in faceCom) is actually implemented and 
    // matchesCommand written in that subclass overides the default 
    // matchesCommand, hence allowing multiple arguments.

    Command faceCom = new Command("face <str...>",this)
      {
	public boolean matchesCommand(String test)
	  {
	    if (test.startsWith("face "))
	      return true;
	    else
	      return false;
	  }
      };
    parse.AddCommand(faceCom);
    parse.AddCommand("show",this);

    display = inDisplay;
  }


  public void clearGraph()
  {
    nodes = new Hashtable();
    edges = new Vector();
    faces = new Vector();
    normals = new Vector();
    retain = new Vector();
    bestVectors = new Vector();
    avgNode = null;
    projectedNodes = new Hashtable();
  }


  public void insertNode(Node inNode)
  {
    nodes.put(inNode.name,inNode);
  }

  public void insertProjectedNode(ProjectedNode inNode)
  {
    projectedNodes.put(inNode.name,inNode);
  }

  public void insertEdge(Edge inEdge)
  {
    edges.addElement(inEdge);
  }

  public void insertFace(Face inFace)
  {
    faces.addElement(inFace);
  }

  public void insertNormal(Node inNormal)
  {
    normals.addElement(inNormal);
  }

  public void averageNode()
  {
    Enumeration e = nodes.elements();
    double x=0,y=0,z=0;
    int sum=0;

    while (e.hasMoreElements())
      {
	Node n = (Node)e.nextElement();
	x += n.x;
	y += n.y;
	z += n.z;
	sum++;
      }

    avgNode = new Node(x/sum,y/sum,z/sum,"avgNode");
    insertNode(avgNode);
    //System.out.println("AverageNode = "+avgNode);
  }

  
  // Normals are defined as the set of unit normal vectors of the planes which contain 
  // more than one element of the set of lines of the object
  public void computeNormals()
  {

    for (int i=0; i < edges.size(); i++) {
      Edge e1 = (Edge)edges.elementAt(i);
      for (int j=0; j < edges.size(); j++) {
	Edge e2 = (Edge)edges.elementAt(j);
	if (!e1.equals(e2)) {
	  if (e1.isCoPlanar(e2)) {
	    Node normal = e1.unitNormal(e2);
	    if (normal != null) {
	      if (!inNormals(normal) && !inNormals(normal.scaleNode(-1))) {
		insertNormal(normal);
	      }
	    }
	  }
	}
      }
    }
    


    //Now that the normals are computed ready to start alogrithm
    Node bv = getBestVector();
    display.setForeground(Color.blue);
    display.append("Using bv : "+bv+"\n");
    Node camera = avgNode.addNode(bv.scaleNode(10));
    mapTo2d(camera);
  }


  public void showView(int view)
  {
    Node bv = (Node)bestVectors.elementAt(view);
    Node camera = avgNode.addNode(bv.scaleNode(10));
    mapTo2d(camera);	
  }


  public void showView(Node n)
  {
    Node camera = avgNode.addNode(n.scaleNode(10));
    mapTo2d(camera);	
  }

  public boolean inBestVectors(Node n)
  {
    for (int i=0; i< bestVectors.size(); i++)
      if (n.equals((Node)bestVectors.elementAt(i)))
	return true;

    return false;
  }

  public boolean inNormals(Node n)
  {
    for (int i=0; i< normals.size(); i++)
      if (n.equals((Node)normals.elementAt(i)))
	return true;

    return false;
  }


  public boolean inRetain(RetainedNode n)
  {
    for (int i=0; i < retain.size(); i++)
      {
	RetainedNode rn = (RetainedNode)retain.elementAt(i);
	if (n.GetFirstNode().equals(rn.GetFirstNode()) && 
	    n.GetSecondNode().equals(rn.GetSecondNode()))
	  return true;
      }
    return false;
  }


  public Node getBestVector()
  {
    int i=0,j=0,k=0;
    Node bestVector = null;
    double min = 0;
    Node t1 = null;
    Node t2 = null;
    Node t3 = null;
    Node nt1 = null;
    Node nt2 = null;
    Node nt3 = null;
    Node pa = null;
    Node pb = null;
    Node pc = null;
    Node pd = null;
    Node pe = null;
    Node pf = null;
    Node sum = null;
    Node diff = null;
    double mindp = 99999;



    if (normals.size() == 1)
      {
	Node n = (Node)normals.elementAt(0);
	RetainedNode rn = new RetainedNode(n,n);
	retain.addElement(rn);
	bestVector = n;
	bestVectors.addElement(bestVector); // note that the last one
					    // is the best one
	return bestVector;
      }

    for (i = 0; i<normals.size(); i++)
      {
	t1 = (Node)normals.elementAt(i);
	for (j = 0; j<normals.size(); j++)
	  {
	    t2 = (Node)normals.elementAt(j);	  
	    if (!t1.equals(t2) && !t1.equals(t2.scaleNode(-1)))
	      {
		sum = t1.addNode(t2);
		diff = t1.subtractNode(t2);
		pa = sum.normalise();
		pb = diff.normalise();

		double padp1 = Math.abs(pa.dotProduct(t1));
		double padp2 = Math.abs(pa.dotProduct(t2));
		double pbdp1 = Math.abs(pb.dotProduct(t1));
		double pbdp2 = Math.abs(pb.dotProduct(t2));
		
		if (compareDoubles(padp1,padp2) && padp1 <= mindp)
		  {
		    RetainedNode rn1 = new RetainedNode(t1,pa);
		    retain.addElement(rn1);
		    if (padp1 < mindp && padp1 > 0.01)
		      mindp = padp1;
		  }

		if (compareDoubles(pbdp1,pbdp2) && padp1 <= mindp)
		  {
		    RetainedNode rn2 = new RetainedNode(t1,pb);
		    retain.addElement(rn2);
		    if (pbdp1 < mindp && pbdp1 > 0.01)
		      mindp = pbdp1;
		  }	      
	      }
	  }
      }


    for (i = 0; i<normals.size(); i++)
      {
	t1 = (Node)normals.elementAt(i);
	for (j = 0; j<normals.size(); j++)
	  {
	    t2 = (Node)normals.elementAt(j);	  
	    for (k = 0; k<normals.size(); k++)
	      {
		t3 = (Node)normals.elementAt(k);
		if (t1.equals(t2) || t1.equals(t2.scaleNode(-1)) || 
		    t1.equals(t3) || t1.equals(t3.scaleNode(-1)) ||
		    t2.equals(t3) || t2.equals(t3.scaleNode(-1)))  //maybe TOO restrictive
		  ;
		else
		  {
		    nt1 = t1.scaleNode(-1.0);
		    nt2 = t2.scaleNode(-1.0);
		    nt3 = t3.scaleNode(-1.0);
		    pc = center(t1,t2,t3);
		    pd = center(nt1,t2,t3);
		    pe = center(t1,nt2,t3);
		    pf = center(t1,t2,nt3);

		    //System.out.println("T1 = "+t1);
		    //System.out.println("pc = "+pc+" , pd = "+pd);
		    //System.out.println("pe = "+pe+" , pf = "+pf);
		    if (pc != null && pd != null && pe != null && pf != null)
		      {
		      
			double pcdp1 = Math.abs(pc.dotProduct(t1));
			double pcdp2 = Math.abs(pc.dotProduct(t2));
			double pcdp3 = Math.abs(pc.dotProduct(t3));
			double pddp1 = Math.abs(pd.dotProduct(t1));
			double pddp2 = Math.abs(pd.dotProduct(t2));
			double pddp3 = Math.abs(pd.dotProduct(t3));
			double pedp1 = Math.abs(pe.dotProduct(t1));
			double pedp2 = Math.abs(pe.dotProduct(t2));
			double pedp3 = Math.abs(pe.dotProduct(t3));
			double pfdp1 = Math.abs(pf.dotProduct(t1));
			double pfdp2 = Math.abs(pf.dotProduct(t2));
			double pfdp3 = Math.abs(pf.dotProduct(t3));

			if (compareDoubles(pcdp1,pcdp2) && compareDoubles(pcdp2,pcdp3) && pcdp1 <= mindp)
			  {
			    RetainedNode rn3 = new RetainedNode(t1,pc);
			    retain.addElement(rn3);
			    if (pcdp1 < mindp && pcdp1 > 0.01)
			      mindp = pcdp1;
			  }


			if (compareDoubles(pddp1,pddp2) && compareDoubles(pddp2,pddp3) && pddp1 <= mindp)
			  {
			    RetainedNode rn4 = new RetainedNode(t1,pd);
			    retain.addElement(rn4);
			    if (pddp1 < mindp && pddp1 > 0.01)
			      mindp = pddp1;
			  }

			if (compareDoubles(pedp1,pedp2) && compareDoubles(pedp2,pedp3) && pedp1 <= mindp)
			  {
			    RetainedNode rn5 = new RetainedNode(t1,pe);
			    retain.addElement(rn5);
			    if (pedp1 < mindp && pedp1 > 0.01)
			      mindp = pedp1;
			  }
			
			if (compareDoubles(pfdp1,pfdp2) && compareDoubles(pfdp2,pfdp3) && pfdp1 <= mindp)
			  {
			    RetainedNode rn6 = new RetainedNode(t1,pf);
			    retain.addElement(rn6);
			    if (pfdp1 < mindp && pfdp1 > 0.01)
			      mindp = pfdp1;
			  }		     
		      }
		  }
	      }
	  }
      }


    //System.out.println("MINDP ====="+mindp);


    for (i = 0; i<retain.size(); i++)
      {
	RetainedNode rn = (RetainedNode)retain.elementAt(i);
	double dp = Math.abs(rn.GetSecondNode().dotProduct(rn.GetFirstNode()));
	
	if (compareDoubles(dp,mindp))
	  {
	    bestVector = rn.GetSecondNode();
	    if (!inBestVectors(bestVector))
	      bestVectors.addElement(bestVector);
	  }
      }


    //System.out.println("Found "+bestVectors.size()+" awesome views");

    bestVect = bestVector;
    return bestVector;
  }

  public int getNumBestVectors()
  {
    return bestVectors.size();
  }


  public boolean compareDoubles(double a, double b)
  {
    double result = a - b;

    if ((result >= -0.0005) && (result <= 0.0005))
      return true;

    return false;
  }

    


  public Node center(Node a, Node b, Node c)
  {
    Node sub1 = a.subtractNode(b);
    Node sub2 = a.subtractNode(c);
    Node result = sub1.crossProduct(sub2);
    if (result.isOrigin())
      {
	System.err.println("WOW!!!!!!!!!!!");
	return null;
      }
    //System.out.println("---->"+result.normalise());
    return result.normalise();
  }


  public void mapTo2d(Node camera)
  {
    Node y = new Node(0,1,0,"yaxis");
    double xrot = 0, yrot = 0, zrot = 0;
    Hashtable rotNodes = new Hashtable();
    Hashtable transNodes = new Hashtable();

    if (avgNode == null)
      {
	System.err.println("Can't compute, avgNode is null");
	return;
      }


    // first get new basis vectors in coordinates defined by the
    // vector from the avgVector to the camera.
    // System.out.println("!!!!"+camera.subtractNode(avgNode));
    Node zprime = camera.subtractNode(avgNode).normalise();
    // Now if zprime is in the direction of the y axis or in the negative y direction
    // have to choose a different zprime
    if (zprime.equals(y) || zprime.equals(y.scaleNode(-1)))
      y = new Node(1,0,0,"xaxis");
    Node xprime = y.crossProduct(zprime).normalise();
    Node yprime = zprime.crossProduct(xprime).normalise();
    // These vectors form the rotation matrix



    // Now have to translate the nodes to new coordinates origin,
    // choose the camera as the new origin ??????? IS THIS RIGHT
    Enumeration e1 = nodes.elements();
    while (e1.hasMoreElements())
      {
	Node n = (Node)e1.nextElement();
	Node tn = n.addNode(camera);
	tn.name = n.name;
	transNodes.put(n.name, tn);
      }
    
    // For all the nodes computed above ,rotate them 
    Enumeration e2 = transNodes.elements();
    while (e2.hasMoreElements())
      {
	Node n = (Node)e2.nextElement();
	xrot = xprime.x*n.x + yprime.x*n.y + zprime.x*n.z;
	yrot = xprime.y*n.x + yprime.y*n.y + zprime.y*n.z;
	zrot = xprime.z*n.x + zprime.x*n.y + zprime.z*n.z;
	Node rn = new Node(xrot,yrot,zrot,n.name);
	rotNodes.put(n.name, rn);
	//System.out.println("Name is "+n.name);
      }
	
    
    // loose the z coord and add to projectedNodes HT
    Enumeration e3 = rotNodes.elements();
    while (e3.hasMoreElements())
      {
	Node n = (Node)e3.nextElement();
	ProjectedNode pn = new ProjectedNode(n.x,n.y,n.name);
	insertProjectedNode(pn);
	//System.out.println("Adding: "+pn);
      }
  }


  public String toString()
  {
    String outBuffer = new String("");
    //for (int i=0;i<faces.size();i++)
    //{
    //outBuffer += "Face #" + i + " = ";
    //outBuffer += faces.elementAt(i).toString() + " ";
    //outBuffer += "\n";
    //}

    outBuffer += "\nEdges:\n";
    for (int i=0;i<edges.size();i++)
      {
	outBuffer += "Edge #" + i + " = ";
	outBuffer += edges.elementAt(i).toString() + " ";
	outBuffer += "\n";
      }

    outBuffer += "\nNodes:\n";
    outBuffer += nodes.toString() + " ";
    outBuffer += "\n";

    /*    outBuffer += "\nNormals:\n";
    for (int i=0;i<normals.size();i++)
      {
	outBuffer += "Normal #" + i + " = ";
	outBuffer += normals.elementAt(i).toString() + " ";
	outBuffer += "\n";
      }
      */

    return outBuffer;
  }
  
  // command execution...

  public void performCommand(String command)
  {
    StringTokenizer toke = new StringTokenizer(command," ,");
    if (command.startsWith("pt"))
      {
	// extract coordinates here, and insert into data struct
	toke.nextToken(); // get 'pt'.
	String name = toke.nextToken();  // get name of the pt.
	double x = Double.valueOf(toke.nextToken()).doubleValue();
	double y = Double.valueOf(toke.nextToken()).doubleValue();
	double z = Double.valueOf(toke.nextToken()).doubleValue();
	Node node = new Node(x,y,z,name);
	insertNode(node);
	display.setForeground(Color.blue);
	display.append("got node: "+name+" x="+x+" y="+y+" z="+z+"\n");
      }
    if (command.startsWith("edge"))
      {
	toke.nextToken(); // get 'edge'.
	String node1Name = toke.nextToken();
	String node2Name = toke.nextToken();
	Node n1 = (Node)nodes.get(node1Name);
	Node n2 = (Node)nodes.get(node2Name);

	if (n1==null || n2==null)
	  {
	    display.setForeground(Color.blue);
	    display.append("Non existant node in HT, edge not possible"+n1+n2+"\n");
	    return;
	  }

	Edge edge = new Edge(n1,n2);
	insertEdge(edge);
	display.setForeground(Color.blue);
	display.append("got edge: "+edge+"\n");
      }
    if (command.startsWith("face"))
      {
	toke.nextToken(); // get 'face'
	Vector faceVector = new Vector();
	while(toke.hasMoreElements())
	  {
	    String nodeName = toke.nextToken();
	    Node n = (Node)nodes.get(nodeName);
	    if (n==null)
	      {
		display.setForeground(Color.blue);
		display.append("Non existant node in HT, face not possible:"+n+"\n");
		return;
	      }
	    faceVector.addElement(n);
	  }	    
	Face face = new Face(faceVector);
	insertFace(face);
	display.setForeground(Color.blue);
	display.append("got face: "+face+"\n");
      }
    if (command.startsWith("show"))
      {
	String toOutput = toString();
	display.setForeground(Color.red);
	display.append(toOutput+"\n");
      }
  }



  // given length, width, height and starting node it makes a 
  // cuboid chosing labels for you.  Note that the length is
  // x axis, width is y, height is z
  public void makeCuboid(Node o, double lx, double wy, double hz)
  {
    double x1 = o.x;
    double y1 = o.y;
    double z1 = o.z;

    Node a = new Node(x1   , y1   , z1   , "a");
    Node b = new Node(x1+lx, y1   , z1   , "b");
    Node c = new Node(x1+lx, y1+wy, z1   , "c");
    Node d = new Node(x1   , y1+wy, z1   , "d"); 
    Node e = new Node(x1   , y1   , z1+hz, "e"); 
    Node f = new Node(x1+lx, y1   , z1+hz, "f"); 
    Node g = new Node(x1+lx, y1+wy, z1+hz, "g"); 
    Node h = new Node(x1   , y1+wy, z1+hz, "h"); 

    insertNode(a);
    insertNode(b);
    insertNode(c);
    insertNode(d);
    insertNode(e);
    insertNode(f);
    insertNode(g);
    insertNode(h);

    insertEdge(new Edge(a,b));
    insertEdge(new Edge(b,c));
    insertEdge(new Edge(c,d));
    insertEdge(new Edge(d,a));
    insertEdge(new Edge(a,e));
    insertEdge(new Edge(b,f));
    insertEdge(new Edge(c,g));
    insertEdge(new Edge(d,h));
    insertEdge(new Edge(e,f));
    insertEdge(new Edge(f,g));
    insertEdge(new Edge(g,h));
    insertEdge(new Edge(h,e));

    Vector f1 = new Vector();
    f1.addElement(a);
    f1.addElement(b);
    f1.addElement(c);
    f1.addElement(d);
    f1.addElement(a);
    Face one = new Face(f1);
    insertFace(one);

    Vector f2 = new Vector();
    f2.addElement(e);
    f2.addElement(f);
    f2.addElement(g);
    f2.addElement(h);
    f2.addElement(e);
    Face two = new Face(f2);
    insertFace(two);

    Vector f3 = new Vector();
    f3.addElement(a);
    f3.addElement(e);
    f3.addElement(f);
    f3.addElement(b);
    f3.addElement(a);    
    Face three = new Face(f3);
    insertFace(three);

    Vector f4 = new Vector();
    f4.addElement(b);
    f4.addElement(c);
    f4.addElement(g);
    f4.addElement(f);
    f4.addElement(b);    
    Face four = new Face(f4);
    insertFace(four);


    Vector f5 = new Vector();
    f5.addElement(c);
    f5.addElement(d);
    f5.addElement(h);
    f5.addElement(g);
    f5.addElement(c);    
    Face five = new Face(f5);
    insertFace(five);

    Vector f6 = new Vector();
    f6.addElement(a);
    f6.addElement(d);
    f6.addElement(h);
    f6.addElement(e);
    f6.addElement(a);    
    Face six = new Face(f6);
    insertFace(six);


    String toOutput = toString();
    System.out.println(toOutput);
  }
}












/*

  // need three nodes from a face to compute the normals
    Node firstNode; 
    Node secondNode;
    Node thirdNode;

    // vec1 is vector b/w first and second nodes
    Node vec1;
    // vec2 is vector b/w second and third nodes
    Node vec2;

    Vector aFace;

    if (faces.size() == 0)
      {
	System.err.println("No faces available for normal computation");
	return;
      }
    
    
    // computes the normals to all the faces
    for (int i = 0; i < faces.size(); i++)
      {
	// elementAt(i) returns an Object, hence the cast to Face which has theFace as a public elt
	aFace = ((Face)faces.elementAt(i)).theFace;
	if (aFace.size() < 3)
	  {
	    System.err.println("This face has less than three nodes, fatal"+faces.elementAt(i));
	    return;
	  }

	firstNode = (Node)aFace.elementAt(0);
	secondNode = (Node)aFace.elementAt(1);
	thirdNode = (Node)aFace.elementAt(2);

	vec1 = secondNode.subtractNode(firstNode);
	vec2 = thirdNode.subtractNode(secondNode);
		
	Node normalNode = (vec1.crossProduct(vec2)).normalise();
	if (normalNode == null)	  
	  System.err.println("Couldn't obtain normal node for face "+aFace);
	else
	  {
	    System.out.println("Normalised cross product:"+normalNode);
	    insertNormal(normalNode);
	  }
      }
  
*/




		      /*		      if ((compareDoubles(Math.abs(pc.dotProduct(t1)), Math.abs(pd.dotProduct(t1)))) &&
			  (compareDoubles(Math.abs(pd.dotProduct(t1)), Math.abs(pe.dotProduct(t1)))) &&
			  (compareDoubles(Math.abs(pe.dotProduct(t1)), Math.abs(pf.dotProduct(t1)))))
			{
			  RetainedNode rn3 = new RetainedNode(t1,pc);
			  RetainedNode rn4 = new RetainedNode(t1,pd);
			  RetainedNode rn5 = new RetainedNode(t1,pe);
			  RetainedNode rn6 = new RetainedNode(t1,pf);
			  if (!inRetain(rn3))
			    retain.addElement(rn3);
			  if (!inRetain(rn4))
			    retain.addElement(rn4);
			  if (!inRetain(rn5))
			    retain.addElement(rn5);
			  if (!inRetain(rn6))
			    retain.addElement(rn6);
			}
			*/
		      





		/*		if (compareDoubles(Math.abs(pa.dotProduct(t1)),Math.abs(pb.dotProduct(t1))))
		  {
		    RetainedNode rn1 = new RetainedNode(t1,pa);
		    RetainedNode rn2 = new RetainedNode(t1,pb);
		    if (!inRetain(rn1))
		      retain.addElement(rn1);
		    if (!inRetain(rn2))
		      retain.addElement(rn2);
		   }
		      */
				
